/*
 * Copyright (c) 1983 Eric P. Allman
 * Copyright (c) 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char sccsid[] = "@(#)recipient.c	5.18 (Berkeley) 6/1/90";
static char  rcsid[] = "@(#)$Id: recipient.c,v 5.18.0.19 1993/12/02 04:11:21 paul Exp $";
#endif /* not lint */

#include "sendmail.h"
#include <sys/stat.h>
#include <pwd.h>
#ifndef S_IEXEC
# define	S_IEXEC		_S_IEXEC
#endif /* !S_IEXEC */
#ifndef S_IWRITE
# define	S_IWRITE	_S_IWRITE
#endif /* !S_IWRITE */

static struct passwd * finduser __P((char *, int *));
static int partialstring __P((const char *, const char *));
static bool writable __P((struct stat *));

/*
**  SENDTOLIST -- Designate a send list.
**
**	The parameter is a comma-separated list of people to send to.
**	This routine arranges to send to all of them.
**
**	Parameters:
**		list -- the send list.
**		ctladdr -- the address template for the person to
**			send to -- effective uid/gid are important.
**			This is typically the alias that caused this
**			expansion.
**		sendq -- a pointer to the head of a queue to put
**			these people into.
**
**	Returns:
**		none
**
**	Side Effects:
**		none.
*/

# define MAXRCRSN	10

void
sendtolist(list, ctladdr, sendq)
	const char *list;
	ADDRESS *ctladdr;
	ADDRESS **sendq;
{
	register const char *p;
	register ADDRESS *al;	/* list of addresses to send to */
	bool firstone;		/* set on first address sent */
	bool selfref;		/* set if this list includes ctladdr */
	char delimiter;		/* the address delimiter */
	char *oldto = CurEnv->e_to;

	if (tTd(25, 1))
	{
		printf("sendto: %s\n   ctladdr=", list);
		printaddr(ctladdr, FALSE);
	}

	/* heuristic to determine old versus new style addresses */
	if (ctladdr == NULL &&
	    (index(list, ',') != NULL || index(list, ';') != NULL ||
	     index(list, '<') != NULL || index(list, '(') != NULL))
		CurEnv->e_flags &= ~EF_OLDSTYLE;
	delimiter = ' ';
	if (!bitset(EF_OLDSTYLE, CurEnv->e_flags) || ctladdr != NULL)
		delimiter = ',';

	firstone = TRUE;
	selfref = FALSE;
	al = NULL;

	for (p = list; *p != '\0'; )
	{
		register ADDRESS *a;
		extern char *DelimChar;		/* defined in prescan */

		/* parse the address */
		while (isspace(*p) || *p == ',')
			p++;
		a = parseaddr((char *)p, (ADDRESS *) NULL, 1, delimiter);
		if (a == NULL)
		{
			/*
			 * Add the address anyway but mark it as bad.
			 * That forces a bounce message to be sent.  This
			 * is for addresses that resolve to the #ERROR mailer.
			 */
			a = (ADDRESS *) xalloc(sizeof *a);
			bzero((char *) a, sizeof *a);
			a->q_flags = QBADADDR|QDONTSEND;
			a->q_paddr = newstr(p);
			a->q_host = newstr("");
			a->q_user = newstr("");
			a->q_mailer = LocalMailer;
		}
		p = DelimChar;
		a->q_next = al;
		a->q_alias = ctladdr;

		/* see if this should be marked as a primary address */
		if (ctladdr == NULL ||
		    (firstone && *p == '\0' && bitset(QPRIMARY, ctladdr->q_flags)))
			a->q_flags |= QPRIMARY;

		/* put on send queue or suppress self-reference */
		if (ctladdr != NULL && sameaddr(ctladdr, a))
			selfref = TRUE;
		else
			al = a;
		firstone = FALSE;
	}

	/* if this alias doesn't include itself, delete ctladdr */
	if (selfref && ctladdr != NULL)
		ctladdr->q_flags |= QSELFREF;

	/* arrange to send to everyone on the local send list */
	while (al != NULL)
	{
		register ADDRESS *a = al;

		al = a->q_next;
		setctladdr(a);
		a = recipient(a, sendq);

		/* arrange to inherit full name */
		if (a->q_fullname == NULL && ctladdr != NULL)
			a->q_fullname = ctladdr->q_fullname;
	}
	CurEnv->e_to = oldto;
}
/*
**  RECIPIENT -- Designate a message recipient
**
**	Saves the named person for future mailing.
**
**	Parameters:
**		a -- the (preparsed) address header for the recipient.
**		sendq -- a pointer to the head of a queue to put the
**			recipient in.  Duplicate supression is done
**			in this queue.
**
**	Returns:
**		The actual address in the queue.  This will be "a" if
**		the address is not a duplicate, else the original address.
**
**	Side Effects:
**		none.
*/

ADDRESS *
recipient(a, sendq)
	register ADDRESS *a;
	register ADDRESS **sendq;
{
	register ADDRESS *q;
	ADDRESS **pq;
	register struct mailer *m;
	register char *p;
	bool quoted = FALSE;		/* set if the addr has a quote bit */
	char buf[MAXNAME];		/* unquoted image of the user name */
	int rcode;

	CurEnv->e_to = a->q_paddr;
	m = a->q_mailer;
	errno = 0;
	if (tTd(26, 1))
	{
		printf("\nrecipient: ");
		printaddr(a, FALSE);
	}

	/* break aliasing loops */
	if (AliasLevel > MAXRCRSN)
	{
		usrerr("aliasing/forwarding loop broken");
		return (a);
	}

	/*
	**  Finish setting up address structure.
	*/

	/* set the queue timeout */
	a->q_timeout = TimeOut;

	/* map user & host to lower case if requested on non-aliases */
	if (a->q_alias == NULL)
		loweraddr(a);

	/* get unquoted user for file, program or user.name check */
	(void) strcpy(buf, a->q_user);
	for (p = buf; *p != '\0' && !quoted; p++)
	{
		if (!isascii(*p) && (*p & 0377) != (SpaceSub & 0377))
			quoted = TRUE;
	}
	stripquotes(buf, TRUE);

	/* do sickly crude mapping for program mailing, etc. */
	if (m == LocalMailer && buf[0] == '|')
	{
		a->q_mailer = m = ProgMailer;
		a->q_user++;
		if (a->q_alias == NULL && !bitset(QGOODUID, a->q_flags) && !ForceMail)
		{
			a->q_flags |= QDONTSEND|QBADADDR;
			usrerr("Cannot mail directly to programs");
		}
	}

	/*
	**  Look up this person in the recipient list.
	**	If they are there already, return, otherwise continue.
	**	If the list is empty, just add it.  Notice the cute
	**	hack to make from addresses suppress things correctly:
	**	the QDONTSEND bit will be set in the send list.
	**	[Please note: the emphasis is on "hack."]
	*/

	for (pq = sendq; (q = *pq) != NULL; pq = &q->q_next)
	{
		if (!ForceMail && sameaddr(q, a))
		{
			if (tTd(26, 1))
			{
				printf("%s in sendq: ", a->q_paddr);
				printaddr(q, FALSE);
			}
			if (!bitset(QDONTSEND, a->q_flags))
				message(Arpa_Info, "duplicate suppressed");
			if (!bitset(QPRIMARY, q->q_flags))
				q->q_flags |= a->q_flags;
			return (q);
		}
	}

	/* add address on list */
	*pq = a;
	a->q_next = NULL;
	CurEnv->e_nrcpts++;

	/*
	**  Alias the name and handle :include: specs.
	*/

	if (m == LocalMailer && !bitset(QDONTSEND, a->q_flags))
	{
		if (strncmp(a->q_user, ":include:", 9) == 0)
		{
			a->q_flags |= QDONTSEND;
			if (a->q_alias == NULL && !bitset(QGOODUID, a->q_flags) && !ForceMail)
			{
				a->q_flags |= QBADADDR;
				usrerr("Cannot mail directly to :include:s");
			}
			else
			{
				message(Arpa_Info, "including file %s", &a->q_user[9]);
				include(&a->q_user[9], " sending", a, sendq);
			}
		}
		else
			alias(a, sendq);
	}

	/*
	**  If the user is local and still being sent, verify that
	**  the address is good.  If it is, try to forward.
	**  If the address is already good, we have a forwarding
	**  loop.  This can be broken by just sending directly to
	**  the user (which is probably correct anyway).
	*/

	if (!bitset(QDONTSEND, a->q_flags) && m == LocalMailer)
	{
		struct stat stb;

		/* see if this is to a file */
		if (buf[0] == '/')
		{
			p = rindex(buf, '/');
			/* check if writable or creatable */
			if (a->q_alias == NULL && !bitset(QGOODUID, a->q_flags) && !ForceMail)
			{
				a->q_flags |= QDONTSEND|QBADADDR;
				usrerr("Cannot mail directly to files");
			}
			else if ((stat(buf, &stb) >= 0) ? (!writable(&stb)) :
			    (*p = '\0', !safefile(buf, getruid(), S_IWRITE|S_IEXEC)))
			{
				a->q_flags |= QBADADDR;
				giveresponse(EX_CANTCREAT, m, CurEnv);
			}
		}
		else
		{
			register struct passwd *pw;

			/*
			 * pw != NULL
			 *	rcode == EX_OK if found via getpwnam()
			 *	rcode == EX_NOUSER if matched via gecos.
			 * pw == NULL
			 *	rcode == EX_NOUSER if not matched at all.
			 */
			pw = finduser(buf, &rcode);
			if (pw == NULL)
			{
				errno = 0;	/* no special error */
				a->q_flags |= QDONTSEND|QBADADDR;
				giveresponse(EX_NOUSER, m, CurEnv);
			}
			else
			{
				char nbuf[MAXNAME];

				if (strcmp(a->q_user, pw->pw_name) != 0)
				{
					a->q_user = newstr(pw->pw_name);
					(void) strcpy(buf, pw->pw_name);
				}
				a->q_home = newstr(pw->pw_dir);
				a->q_uid = pw->pw_uid;
				a->q_gid = pw->pw_gid;
				a->q_flags |= QGOODUID;
				buildfname(pw->pw_gecos, pw->pw_name, nbuf);
				if (nbuf[0] != '\0')
					a->q_fullname = newstr(nbuf);

				/* do extra alias lookup if found via gecos */
				if (rcode != EX_OK)
					alias(a, sendq);

				/* forward if still being sent */
				if (!bitset(QDONTSEND, a->q_flags))
				{
					if (!quoted)
						forward(a, sendq);
				}
			}
		}
	}
	return (a);
}
/*
**  FINDUSER -- find the password entry for a user.
**
**	This looks a lot like getpwnam, except that it may want to
**	do some fancier pattern matching in /etc/passwd.
**
**	This routine contains most of the time of many sendmail runs.
**	It deserves to be optimized.
**
**	Parameters:
**		name -- the name to match against.
**
**	Returns:
**		A pointer to a pw struct.
**		NULL if name is unknown or ambiguous.
**
**	Side Effects:
**		may modify name.
*/

#define WORST_MATCH	-2		/* even worse than no match */
#define NO_UID		-999		/* any "impossible" uid will do */

static struct passwd *
finduser(name, rcode)
	char *name;
	int *rcode;
{
	register struct passwd *pw;
#ifdef FUZZY
	int best_match = WORST_MATCH;
	int best_uid = NO_UID;
#endif /* FUZZY */

	errno = 0;

	/* first try name as given */
	if ((pw = getpwnam(name)) == NULL)
	{
		/* try again as lower-case */
		char *lowname = newstr(name);
		makelower(lowname);
		pw = getpwnam(lowname);
		free(lowname);
	}
	(pw == NULL) ? (*rcode = EX_NOUSER) : (*rcode = EX_OK);

	if (tTd(26, 6))
		printf("%s password entry for \"%s\"\n",
		       pw ? "found" : "can't find", name);

#ifndef FUZZY
	return (pw);
#else /* FUZZY */
	if (pw != NULL)
		return (pw);

	if (tTd(26, 6))
		printf("looking for partial match to \"%s\"\n", name);
	(void) setpwent();
	while ((pw = getpwent()) != NULL)
	{
		char buf[MAXNAME];
		register int this_match;

		if (strcasecmp(pw->pw_name, name) == 0)
		{
		    if (tTd(26, 6))
			printf("found password entry for \"%s\" as \"%s\"\n",
				   name, pw->pw_name);
		    return (pw);
		}
		buildfname(pw->pw_gecos, pw->pw_name, buf);
		this_match = partialstring(buf, name);
		if (tTd(26, 6) && this_match >= 0)
			printf("matched on level %d with \"%s\"\n",
			       this_match, buf);
		if (this_match < best_match)
			continue;
		else if (this_match > best_match)
		{
			best_match = this_match;
			best_uid = pw->pw_uid;
		}
		else if (best_uid != pw->pw_uid)
			best_uid = NO_UID;
	}
	if (tTd(26, 6))
	{
		if (best_match == WORST_MATCH)
			printf("no match, failing...\n");
		else if (best_uid == NO_UID)
			printf("ambiguous match, failing...\n");
		else
			printf("succeding on level %d...\n",
			       best_match);
	}
	if (best_uid == NO_UID)
		return (NULL);

	pw = getpwuid(best_uid);
			message(Arpa_Info, "sending to login name %s", pw->pw_name);
			return (pw);
#endif /* !FUZZY */
}
/*
**  PARTIALSTRING -- is one string of words contained by another?
**
**	See if one string of words can be found as part of
**	another string of words.  All substrings delimited by
**	one or more non-alphanumeric characters are considered
**	"words", and a partial match is such that all the words
**	of the pattern string are either full prefixes
**	of the target string.  Upper or lower case letters are
**	considered equal.
**
**	Parameters:
**		target -- target string
**		pattern -- pattern string
**
**	Returns:
**		The number of fully matched words, or -1 if none.
**
**	Side Effects:
**		None.
**
*/

#ifdef FUZZY
static int
partialstring(target, pattern)
	const char *target, *pattern;
{
	register char *t, *p, *q;
	int full_words = 0;

	/* skip initial delimiters */
	for (t = (char *)target; *t != '\0' && !isalnum(*t); t++)
		;
	for (p = (char *)pattern; *p != '\0' && !isalnum(*p); p++)
		;
	q = p;

	while (*t != '\0' && *p != '\0')
	{
		/*
		 * if at end of pattern word, find next, remember it,
		 * and eat the current target word
		 */
		if (!isalnum(*p))
		{
			while (*p != '\0' && !isalnum(*p))
				p++;
			if (*p == '\0')
				continue;
			q = p;
			if (!isalnum(*t))
				full_words++;
			while (*t != '\0' && isalnum(*t))
				t++;
			while (*t != '\0' && !isalnum(*t))
				t++;
			continue;
		}

		/*
		 * if match, advance both pointers
		 */
		if ((isupper(*t) ? tolower(*t) : *t) ==
		    (isupper(*p) ? tolower(*p) : *p))
		{
			t++, p++;
			continue;
		}

		/*
		 * if no match, backtrack to last unmatched pattern word and
		 * eat current target word.
		 */
		p = q;
		while (*t != '\0' && isalnum(*t))
			t++;
		while (*t != '\0' && !isalnum(*t))
			t++;
	}

	/*
	 * now, the pattern should be fully consumed if there was a match
	 */
	if (*p == '\0')
		return (isalnum(*t) ? full_words : full_words + 1);
	else
		return (-1);
}
#endif /* FUZZY */
/*
**  WRITABLE -- predicate returning if the file is writable.
**
**	This routine must duplicate the algorithm in sys/fio.c.
**	Unfortunately, we cannot use the access call since we
**	won't necessarily be the real uid when we try to
**	actually open the file.
**
**	Notice that ANY file with ANY execute bit is automatically
**	not writable.  This is also enforced by mailfile.
**
**	Parameters:
**		s -- pointer to a stat struct for the file.
**
**	Returns:
**		TRUE -- if we will be able to write this file.
**		FALSE -- if we cannot write this file.
**
**	Side Effects:
**		none.
*/

static bool
writable(s)
	register struct stat *s;
{
	int euid, egid;
	int bits;

	if (bitset(0111, s->st_mode))
		return (FALSE);
	euid = getruid();
	egid = getrgid();
	if (geteuid() == 0)
	{
		if (bitset(S_ISUID, s->st_mode))
			euid = s->st_uid;
		if (bitset(S_ISGID, s->st_mode))
			egid = s->st_gid;
	}

	if (euid == 0)
		return (TRUE);
	bits = S_IWRITE;
	if (euid != s->st_uid)
	{
		bits >>= 3;
		if (egid != s->st_gid)
			bits >>= 3;
	}
	return ((s->st_mode & bits) != 0);
}
/*
**  INCLUDE -- handle :include: specification.
**
**	Parameters:
**		fname -- filename to include.
**		msg -- message to print in verbose mode.
**		ctladdr -- address template to use to fill in these
**			addresses -- effective user/group id are
**			the important things.
**		sendq -- a pointer to the head of the send queue
**			to put these addresses in.
**
**	Returns:
**		none.
**
**	Side Effects:
**		reads the :include: file and sends to everyone
**		listed in that file.
*/

void
include(fname, msg, ctladdr, sendq)
	const char *fname, *msg;
	ADDRESS *ctladdr;
	ADDRESS **sendq;
{
	char buf[MAXLINE];
	register FILE *fp;
	char *oldto = CurEnv->e_to;
	char *oldfilename = FileName;
	int oldlinenumber = LineNumber;
	bool included = FALSE;

	fp = fopen(fname, "r");
	if (fp == NULL)
	{
		usrerr("Cannot open %s", fname);
		return;
	}
	if (getctladdr(ctladdr) == NULL)
	{
		struct stat st;

		if (fstat(fileno(fp), &st) < 0)
			syserr("Cannot fstat %s!", fname);
		ctladdr->q_uid = st.st_uid;
		ctladdr->q_gid = st.st_gid;
		ctladdr->q_flags |= QGOODUID;
	}

	/* sendtolist() detects possible self-refs in any line of this file */
	ctladdr->q_flags &= ~QSELFREF;

	/* read the file -- each line is a comma-separated list. */
	FileName = (char *)fname;
	LineNumber = 0;
	while (fgets(buf, sizeof buf, fp) != NULL)
	{
		register char *p = index(buf, '\n');

		LineNumber++;
		if (p != NULL)
			*p = '\0';

		/* ignore null lines and # comment lines */
		if ((buf[0] == '\0') || (buf[0] == '#'))
			continue;
		CurEnv->e_to = oldto;
		message(Arpa_Info, "%s to %s", msg, buf);
		AliasLevel++;
		sendtolist(buf, ctladdr, sendq);
		included = TRUE;
		AliasLevel--;
	}

	if (included && !bitset(QSELFREF, ctladdr->q_flags))
		ctladdr->q_flags |= QDONTSEND;
	(void) fclose(fp);
	FileName = oldfilename;
	LineNumber = oldlinenumber;
}
/*
**  SENDTOARGV -- send to an argument vector.
**
**	Parameters:
**		argv -- argument vector to send to.
**
**	Returns:
**		none.
**
**	Side Effects:
**		puts all addresses on the argument vector onto the
**			send queue.
*/

void
sendtoargv(argv)
	register char **argv;
{
	register char *p;

	while ((p = *argv++) != NULL)
	{
		if (argv[0] != NULL && argv[1] != NULL && !strcasecmp(argv[0], "at"))
		{
			char nbuf[MAXNAME];

			if (strlen(p) + strlen(argv[1]) + 2 > sizeof nbuf)
				usrerr("address overflow");
			else
			{
				(void) strcpy(nbuf, p);
				(void) strcat(nbuf, "@");
				(void) strcat(nbuf, argv[1]);
				p = newstr(nbuf);
				argv += 2;
			}
		}
		sendtolist(p, (ADDRESS *) NULL, &CurEnv->e_sendqueue);
	}
}
/*
**  GETCTLADDR -- get controlling address from an address header.
**
**	If none, get one corresponding to the effective userid.
**
**	Parameters:
**		a -- the address to find the controller of.
**
**	Returns:
**		the controlling address.
**
**	Side Effects:
**		none.
*/

ADDRESS *
getctladdr(a)
	register ADDRESS *a;
{
	while (a != NULL && !bitset(QGOODUID, a->q_flags))
		a = a->q_alias;
	return (a);
}
