/*
 * Copyright (c) 1983, 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * SYSLOG -- print message on log file
 *
 * This routine looks a lot like printf, except that it outputs to the
 * log file instead of the standard output.  Also:
 *	adds a timestamp,
 *	prints the module name in front of the message,
 *	has some other formatting types (or will sometime),
 *	adds a newline on the end of the message.
 *
 * The output of this routine is intended to be read by syslogd(8).
 *
 * Author: Eric Allman
 * Modified to use UNIX domain IPC by Ralph Campbell
 * modified for System V by Bruce Lilly
 */

#include "sendmail.h"
#ifdef LACK_SYSLOG

#ifndef lint
static char sccsid[] = "@(#)syslog.c	5.20 (Berkeley) 1/19/89";
static char  rcsid[] = "@(#)$Id: syslog.c,v 5.20.0.7 1992/02/27 16:40:14 paul Exp $";
#endif /* not lint */

# ifdef SYSV
#  include <string.h>
# else /* !SYSV */
#  include <strings.h>
# endif /* SYSV */
# ifndef SYSLOG_NOIPC
#  include <netdb.h>
# endif /* !SYSLOG_NOIPC */

# ifdef __STDC__
void	closelog();
# else /* !__STDC__ */
void	closelog(), vsyslog();
# endif /* __STDC__ */

# define	LOGNAME	"/dev/log"
# define	CONSOLE	"/dev/console"

static int	LogFile = -1;		/* fd for log */
static int	connected;		/* have done connect */
static int	LogStat = 0;		/* status bits, set by openlog() */
static char	*LogTag = "syslog";	/* string to tag the entry with */
static int	LogFacility = LOG_USER;	/* default facility code */

void		openlog();

# ifndef LOG_FAC
#  define LOG_FAC(p)	(((p) & LOG_FACMASK) >> 3)	/* facility of pri */
# endif /* !LOG_FAC */

# ifndef LOG_PRI
#  define LOG_PRI(p)	((p) & LOG_PRIMASK)	/* extract priority */
# endif /* !LOG_PRI */

# ifndef __STDC__
void
syslog(pri, fmt, args)
	int pri, args;
	char *fmt;
{
	vsyslog(pri, fmt, &args);
}

/*VARARGS2*/
void
vsyslog(pri, fmt, va_alist)
	int pri;
	register char *fmt;
	va_dcl
# else /* __STDC__ */

void
syslog(int pri, char *fmt, ...)
# endif /* !__STDC__ */
{
  	va_list ap;
	register int cnt;
	register char *p;
	TIME_TYPE now;
	int pid, saved_errno;
	char tbuf[2048], fmt_cpy[1024];
	unsigned int more_time;

# ifdef __STDC__
	va_start(ap, fmt);
# else /* !__STDC__ */
	va_start(ap);
# endif /* __STDC__ */
	saved_errno = errno;

	/* see if we should just throw out this message */
	if ((u_int)LOG_FAC(pri) >= LOG_NFACILITIES ||
	    !LOG_MASK(LOG_PRI(pri)) || (pri &~ (LOG_PRIMASK|LOG_FACMASK)))
	{
		va_end(ap);
		return;
	}
	if (LogFile < 0 || !connected)
		openlog(LogTag, LogStat | LOG_NDELAY, 0);

	/* set default facility if none specified */
	if ((pri & LOG_FACMASK) == 0)
		pri |= LogFacility;

	/* build the message */
	(void) time(&now);
	(void) sprintf(tbuf, "<%d>%.15s ", pri, ctime(&now) + 4);
	for (p = tbuf; *p; ++p)
		;
	if (LogTag)
	{
		(void) strcpy(p, LogTag);
		for (; *p; ++p)
			;
	}
	if (LogStat & LOG_PID)
	{
		(void) sprintf(p, "[%d]", getpid());
		for (; *p; ++p)
			;
	}
	if (LogTag)
	{
		*p++ = ':';
		*p++ = ' ';
	}

	/* substitute error message for %m */
	{
		register char ch, *t1, *t2;

		for (t1 = fmt_cpy; ch = *fmt; ++fmt)
		{
			if (ch == '%' && fmt[1] == 'm')
			{
				fmt++;
				for (t2 = errstring(saved_errno);
				    *t1 = *t2++; ++t1)
					;
			}
			else
				*t1++ = ch;
		}
		*t1 = '\0';
	}

	(void) vsprintf(p, fmt_cpy, ap);
	va_end(ap);

# ifdef SYSLOG_NOIPC
	if (strchr(tbuf, '\n' ) == NULL)
		strcat(tbuf, "\n");
# endif /* SYSLOG_NOIPC */

	/* output the message to the local logger */
# ifdef SYSLOG_NOIPC
	if (write(LogFile, tbuf, cnt = strlen(tbuf)) >= 0 ||
# else /* !SYSLOG_NOIPC */
	if (send(LogFile, tbuf, cnt = strlen(tbuf), 0) >= 0 ||
# endif /* SYSLOG_NOIPC */
	    !(LogStat & LOG_CONS))
		return;

	/* output the message to the console */
	pid = vfork();
	if (pid == -1)
		return;
	if (pid == 0)
	{
		int fd;

		(void) signal(SIGALRM, SIG_DFL);
# ifndef SYSV
		sigsetmask((long) ~sigmask(SIGALRM));
# endif /* !SYSV */
		more_time = 5;
		while (more_time = alarm(more_time))
			;
		if ((fd = open(CONSOLE, O_WRONLY, 0)) < 0)
			return;
		(void) alarm((u_int)0);
		(void) strcat(tbuf, "\r");
		p = index(tbuf, '>') + 1;
		(void) write(fd, p, cnt + 1 - (p - tbuf));
		(void) close(fd);
		_exit(0);
	}
	if (!(LogStat & LOG_NOWAIT))
		while ((cnt = wait((int *)0)) > 0 && cnt != pid);
}

# ifndef SYSLOG_NOIPC
static struct sockaddr SyslogAddr;	/* AF_UNIX address of local logger */
static struct sockaddr_in SyslogPort;	/* AF_INET address of local logger */
# endif /* !SYSLOG_NOIPC */

/*
 * OPENLOG -- open system log
 */
void
openlog(ident, logstat, logfac)
	char *ident;
	int logstat, logfac;
{
	if (ident != NULL)
		LogTag = ident;
	LogStat = logstat;
	if (logfac != 0 && (logfac &~ LOG_FACMASK) == 0)
		LogFacility = logfac;
	if (LogFile == -1)
	{
# ifdef SYSLOG_NOIPC
		LogFile = open(LOGNAME, O_CREAT | O_WRONLY | O_APPEND, 0660);
# else /* !SYSLOG_NOIPC */
		SyslogAddr.sa_family = AF_UNIX;
		strncpy(SyslogAddr.sa_data, LOGNAME, sizeof SyslogAddr.sa_data);
		if (LogStat & LOG_NDELAY)
		{
			LogFile = socket(AF_UNIX, SOCK_DGRAM, 0);
			fcntl(LogFile, F_SETFD, 1);
		}
# endif /* SYSLOG_NOIPC */
	}
	if (LogFile != -1 && !connected &&
# ifdef SYSLOG_NOIPC
	    /* Owner = daemon Group = daemon */
	    fchown(LogFile, 1, 1) != -1)
# else /* !SYSLOG_NOIPC */
	    connect(LogFile, &SyslogAddr, sizeof(SyslogAddr)) != -1)
# endif /* SYSLOG_NOIPC */
	{
		connected = 1;
		return;
	}
	else if (LogFile != -1)
		(void) close(LogFile);
# ifdef SYSLOG_NOIPC
	LogStat |= LOG_CONS;
# else /* !SYSLOG_NOIPC */
	if (LogFile == -1)
	{
		struct	servent	*se;
		struct	hostent	*host = 0;
		char	hbuf[32];

		if (gethostname(hbuf, 31) == 0)
		{
			strtok(hbuf, ".");
			host = gethostbyname(hbuf);
		}
		if (host == NULL)
			if ((host = gethostbyname("localhost")) == NULL)
				return;
		if ((se = getservbyname("syslog", "udp")) == NULL)
			return;
		SyslogPort.sin_port = se->s_port;
		SyslogPort.sin_family = AF_INET;
		if ((LogFile = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
			return;
		(void) memcpy((caddr_t)&SyslogPort.sin_addr,
		    (caddr_t)host->h_addr, host->h_length);
	}
	if (LogFile != -1 && !connected &&
	    connect(LogFile, (struct sockaddr *)&SyslogPort, (int)sizeof(SyslogPort)) != -1)
	{
		connected = 1;
		return;
	}
	else
	{
		LogStat |= LOG_CONS;
		if (LogFile != -1)
			(void) close(LogFile);
	}
# endif /* SYSLOG_NOIPC */
}

/*
 * CLOSELOG -- close the system log
 */
void
closelog()
{
	(void) close(LogFile);
	LogFile = -1;
	connected = 0;
}

static int	LogMask = 0xff;		/* mask of priorities to be logged */
/*
 * SETLOGMASK -- set the log mask level
 */
int
setlogmask(pmask)
int	pmask;
{
	int omask;

	omask = LogMask;
	if (pmask != 0)
		LogMask = pmask;
	return (omask);
}
#endif /* LACK_SYSLOG */
